
/*****************************************************************************/
/*                                                                           */
/*  THE NRCONV NURSE ROSTERING TO XESTT CONVERTER                            */
/*  COPYRIGHT (C) 2016, Jeffrey H. Kingston                                  */
/*                                                                           */
/*  Jeffrey H. Kingston (jeff@it.usyd.edu.au)                                */
/*  School of Information Technologies                                       */
/*  The University of Sydney 2006                                            */
/*  AUSTRALIA                                                                */
/*                                                                           */
/*  This program is free software; you can redistribute it and/or modify     */
/*  it under the terms of the GNU General Public License as published by     */
/*  the Free Software Foundation; either Version 3, or (at your option)      */
/*  any later version.                                                       */
/*                                                                           */
/*  This program is distributed in the hope that it will be useful,          */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/*  GNU General Public License for more details.                             */
/*                                                                           */
/*  You should have received a copy of the GNU General Public License        */
/*  along with this program; if not, write to the Free Software              */
/*  Foundation, Inc., 59 Temple Place, Suite 330, Boston MA 02111-1307 USA   */
/*                                                                           */
/*  FILE:         nrc_archive_metadata.c                                     */
/*  DESCRIPTION:  Archive metadata                                           */
/*                                                                           */
/*****************************************************************************/
#include <stdarg.h>
#include "nrc_interns.h"


/*****************************************************************************/
/*                                                                           */
/*  NRC_ARCHIVE_METADATA - archive metadata                                  */
/*                                                                           */
/*****************************************************************************/

struct nrc_archive_metadata_rec {
  char			*name;			/* name                      */
  char			*contributor;		/* contributor               */
  char			*date;			/* date                      */
  char			*description;		/* description               */
  char			*remarks;		/* remarks (optional)        */
};


/*****************************************************************************/
/*                                                                           */
/*  Submodule "construction and query"                                       */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/*                                                                           */
/*  NRC_ARCHIVE_METADATA NrcArchiveMetaDataMake(char *name,                  */
/*    char *contributor, char *date, char *description, char *remarks)       */
/*                                                                           */
/*  Make archive metadata with these attributes.                             */
/*                                                                           */
/*****************************************************************************/

NRC_ARCHIVE_METADATA NrcArchiveMetaDataMake(char *name, char *contributor,
  char *date, char *description, char *remarks)
{
  NRC_ARCHIVE_METADATA res;
  MMake(res);
  res->name = name;
  res->contributor = contributor;
  res->date = date;
  res->description = description;
  res->remarks = remarks;
  return res;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcArchiveMetaDataName(NRC_ARCHIVE_METADATA md)                    */
/*                                                                           */
/*  Return the Name attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *NrcArchiveMetaDataName(NRC_ARCHIVE_METADATA md)
{
  return md->name;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcArchiveMetaDataContributor(NRC_ARCHIVE_METADATA md)             */
/*                                                                           */
/*  Return the Contributor attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *NrcArchiveMetaDataContributor(NRC_ARCHIVE_METADATA md)
{
  return md->contributor;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcArchiveMetaDataDate(NRC_ARCHIVE_METADATA md)                    */
/*                                                                           */
/*  Return the Date attribute of md.                                         */
/*                                                                           */
/*****************************************************************************/

char *NrcArchiveMetaDataDate(NRC_ARCHIVE_METADATA md)
{
  return md->date;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcArchiveMetaDataDescription(NRC_ARCHIVE_METADATA md)             */
/*                                                                           */
/*  Return the Description attribute of md.                                  */
/*                                                                           */
/*****************************************************************************/

char *NrcArchiveMetaDataDescription(NRC_ARCHIVE_METADATA md)
{
  return md->description;
}


/*****************************************************************************/
/*                                                                           */
/*  char *NrcArchiveMetaDataRemarks(NRC_ARCHIVE_METADATA md)                 */
/*                                                                           */
/*  Return the optional Remarks attribute of md.                             */
/*                                                                           */
/*****************************************************************************/

char *NrcArchiveMetaDataRemarks(NRC_ARCHIVE_METADATA md)
{
  return md->remarks;
}
